from typing import List

from .colony_state import ColonyState
from .exception.colony_exception import ColonyException
from ..abstract.gol_interface import GameOfLife as GOLAbstract
from ..gol.field import Field


class Colony(GOLAbstract):

    def __init__(self, gol_engine: GOLAbstract):
        self.gol = gol_engine
        self._field_list: List[Field] = []
        self._field_hash_list: List[int] = []
        self._state: str = ColonyState.ALIVE
        self._add_field(self.gol.field())

    def field(self) -> Field:
        return self.gol.field()

    def next_generation(self) -> Field:
        ng: Field = self.gol.next_generation()
        self._add_field(ng)
        return ng

    def init_field(self, x: int, y: int, start_generation: Field) -> None:
        raise NotImplementedError

    def _add_field(self, field: Field) -> None:
        hash_value: int = self.generate_hash(field.__str__())
        self._field_list.append(field)
        self._field_hash_list.append(hash_value)
        self._test_if_not_dead()
        self._test_if_not_repeated()

    def _is_dead(self) -> bool:
        if '1' in self.gol.field().__str__():
            return False
        else:
            return True

    def _is_repeated(self, hash_value: int) -> bool:
        if hash_value in self._field_hash_list[:-1]:
            return True
        else:
            return False

    def _test_if_not_dead(self):
        if self._is_dead():
            self._state = ColonyState.DEAD
            raise ColonyException.is_dead()

    def _test_if_not_repeated(self):
        if self._is_repeated(self._field_hash_list[-1]):
            self._state = ColonyState.LOOPED
            raise ColonyException.is_repeated()

    @staticmethod
    def generate_hash(field: Field) -> int:
        return hash(field)



from typing import List

from ...gol.field import Field
from ...gol.point import Point


def input_field(string: str, width: int) -> Field:
    row: List = []
    field_list: List[List] = []

    for char in string:
        if char == '0' or char == '1':
            row.append(int(char))
        if len(row) == width:
            field_list.append(row)
            row = []

    if len(row) > 0:
        field_list.append(row)

    field = Field(Point(len(field_list[0]), len(field_list)))
    field.init(field_list)
    return field





class ColonyException(Exception):

    DEAD: int = 1
    LOOP: int = 2

    def __init__(self, code: int = 0, message: str = ''):
        self.code = code
        self.message = message

    @classmethod
    def is_dead(cls):
        return ColonyException(cls.DEAD, 'The colony seems to be DEAD :(')

    @classmethod
    def is_repeated(cls):
        return ColonyException(cls.LOOP, 'The colony seems to repeat its state')





from .point import Point
from typing import List


class Field:

    _geometry: Point
    _field: List[List[int]]

    def __init__(self, geometry: Point):
        self._geometry = geometry
        self._create_field()

    def geometry(self) -> Point:
        return self._geometry

    def init(self, field: List[List[int]]) -> None:
        x = y = 0
        for i in field:
            for j in i:
                self.set(x, y, bool(j))
                x += 1
            y += 1
            x = 0

    def state(self, x: int, y: int) -> bool:
        result = False
        if (0 <= x < self._geometry.x) and (0 <= y < self._geometry.y):
            result = self._field[y][x] != 0
        return result

    def state_point(self, point: Point) -> bool:
        return self.state(point.x, point.y)

    def set(self, x: int, y: int, value: bool) -> None:
        if 0 <= x < self._geometry.x and 0 <= y < self._geometry.y:
            self._field[y][x] = int(value == True)

    def set_point(self, point: Point, value: bool) -> None:
        return self.set(point.x, point.y, value)

    def width(self) -> int:
        return len(self._field[0])

    def _create_field(self) -> None:
        self._field = [
            [0 for i in range(self._geometry.x)] for j in range(self._geometry.y)
        ]

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self._geometry.x * self._geometry.y:
            y = self.n // self._geometry.x
            if y > 0:
                x = self.n - (self._geometry.x * y)
            else:
                x = self.n
            result = Point(x, y)
            self.n += 1
        else:
            raise StopIteration
        return result

    def __str__(self):
        result: str = ''
        for point in self:
            if self.state_point(point):
                char = '1'
            else:
                char = '0'
            result += char
        return result




from ..rules_interface import Rules as RulesAbstract
from ..field import Field
from ..point import Point


class ConwayRules(RulesAbstract):

    NAME = 'CONWAY'

    def name(self) -> str:
        return self.NAME

    def calculate(self, field: Field, point: Point) -> bool:
        result = False

        environment_points = self._environment(point)
        environment_life_cells = self._get_states(field, environment_points)
        # environment_life_cells = self._sum_life_cells(environment_states)

        if field.state_point(point):
            if environment_life_cells in (2, 3):
                result = True
        else:
            if environment_life_cells == 3:
                result = True

        return result

    @staticmethod
    def _environment(point: Point) -> list:
        result = []
        for x in range(point.x - 1, point.x + 2):
            for y in range(point.y - 1, point.y + 2):
                if x != point.x or y != point.y:
                    result.append(Point(x,y))
        return result

    @staticmethod
    def _get_states(field: Field, points: list) -> int:
        sum: int = 0
        for p in points:
            if field.state_point(p):
                sum += 1
        return sum



import os

from .arguments import parse
from ..colony.colony import Colony
from ..colony.exception.colony_exception import ColonyException
from ..gol.arguments import Arguments as GolArguments
from .start_field.only_string import input_field as field_from_string
from .start_field.string_with_width import input_field as field_from_string_with_width
from .start_field.from_file import input_field as field_from_file
from ..gol.factory import Factory
from ..gol.field import Field


def create_gol_arguments(args: object) -> GolArguments:
    if args.file_start_generation:
        field = field_from_file(args.file_start_generation, args.file_false, args.file_true)
    elif (args.width > 0):
        field = field_from_string_with_width(args.start_generation, args.width)
    elif args.start_generation != '':
        field = field_from_string(args.start_generation)
    else:
        print("Can't determine the start generation - please use --help to find how to use this app")
        exit(1)

    width: int = field.width()
    start_generation: str = field.__str__()

    return GolArguments(start_generation, width)


def show_generation(generation_number: int, generation: Field) -> None:
    print(f" >> gen: {generation_number}")
    for point in generation:
        value = generation.state_point(point)
        if value:
            # char = chr(9634)
            # char = ' ' + chr(9632)
            char = 'O '
        else:
            # char = chr(1468) + ' '
            char = '. '
        if point.x == generation.geometry().x - 1:
            print(char)
        else:
            print(char, end='')


def run():

    args = parse()

    gol_arguments = create_gol_arguments(args)
    gol = Factory.create_from_arguments(gol_arguments)
    colony = Colony(gol)

    try:
        for i in range(0, args.generations + 1):

            if args.clear:
                os.system('clear')

            if args.quiet:
                print('.', end='')
            else:
                show_generation(i, colony.field())

            if args.interactive:
                input("Press Enter to continue...")

            if i < args.generations:
                    colony.next_generation()

    except ColonyException as exception:
        print()
        print(exception.message)
        print(f"generation: {i}")


if __name__ == '__main__':
    run()




from abc import ABC
from abc import abstractmethod
from .field import Field
from .point import Point


class Rules(ABC):

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def calculate(self, field: Field, point: Point) -> bool:
        pass




from typing import List
from .point import Point
from .field import Field
from .rules_interface import Rules
from ..abstract.gol_interface import GameOfLife as GOLAbstract


class GameOfLife(GOLAbstract):

    _field: Field
    _rules: Rules

    def __init__(self, width: int, height: int, rules: Rules):
        self._field = Field(Point(width, height))
        self._rules = rules

    def field(self) -> Field:
        return self._field

    def next_generation(self) -> Field:
        ng = Field(self._field.geometry())
        for p in self._field:
            ng.set_point(p, self._rules.calculate(self._field, p))
        self._field = ng
        return self._field

    def init_field(self, x: int, y: int, start_generation: Field) -> None:
        self._field = start_generation




from abc import ABC
from abc import abstractmethod
from ..gol.field import Field
from ..gol.point import Point


class GameOfLife(ABC):

    @abstractmethod
    def field(self) -> Field:
        pass

    @abstractmethod
    def next_generation(self) -> Field:
        pass

    @abstractmethod
    def init_field(self, x: int, y: int, start_generation: Field) -> None:
        pass



from typing import List

from ...gol.field import Field
from ...gol.point import Point


def input_field(file: str, char_false: str, char_true: str) -> Field:
    row: List = []
    field_list: List[List] = []

    with open(file) as f:
        lines = f.readlines()

    for line in lines:
        for char in line.strip():
            if char == '0':
                row.append(0)
            else:
                row.append(1)
        field_list.append(row)
        row = []

    field = Field(Point(len(field_list[0]), len(field_list)))
    field.init(field_list)
    return field



from simple_value_object import ValueObject, invariant


class Arguments(ValueObject):

    start_generation: str = ''
    width: int = 0

    def __init__(self, start_generation, width=0):
        pass

    @invariant
    def start_generation_is_str(self, instance):
        return type(instance.start_generation) == str

    @invariant
    def width_is_int(self, instance):
        return type(instance.width) == int



import pytest

from ..colony import Colony
from ..exception.colony_exception import ColonyException
from ...gol.factory import Factory
from ...gol.field import Field
from ...gol.point import Point


def test_empty():
    with pytest.raises(ColonyException) as colony_exception:
        gol = Factory.create(1, 1)
        Colony(gol)
    assert colony_exception.value.code == ColonyException.DEAD


def test_not_empty():
    field = Field(Point(1, 1))
    field.set(0, 0, True)
    gol = Factory.create_with_start_generation(field)
    colony = Colony(gol)
    assert type(colony) == Colony


def test_init_not_implemented():
    with pytest.raises(NotImplementedError):
        field = Field(Point(1, 1))
        field.set(0, 0, True)
        gol = Factory.create_with_start_generation(field)
        colony = Colony(gol)
        colony.init_field(0, 0, field)


def test_field():
    field = Field(Point(1, 1))
    field.set(0, 0, True)
    gol = Factory.create_with_start_generation(field)
    colony = Colony(gol)
    assert field == colony.field()


def test_next_generation():
    with pytest.raises(ColonyException) as colony_exception:
        field = Field(Point(1, 1))
        field.set(0, 0, True)
        gol = Factory.create_with_start_generation(field)
        colony = Colony(gol)
        colony.next_generation()
    assert colony_exception.value.code == ColonyException.DEAD


def test_next_generation_repeated():
    with pytest.raises(ColonyException) as colony_exception:
        field = Field(Point(3, 3))
        field.set(0, 1, True)
        field.set(1, 1, True)
        field.set(2, 1, True)
        gol = Factory.create_with_start_generation(field)
        colony = Colony(gol)
        colony.next_generation()
        colony.next_generation()
    assert colony_exception.value.code == ColonyException.LOOP



from .arguments import Arguments
from .field import Field
from ..abstract.gol_interface import GameOfLife as GOLInterface
from .api import GameOfLife
from .rules.conway import ConwayRules
from ..cli.start_field.string_with_width import input_field


class Factory:

    @staticmethod
    def create(width: int, height: int) -> GOLInterface:
        rules = ConwayRules()
        return GameOfLife(width, height, rules)

    @staticmethod
    def create_with_start_generation(start_generation: Field) -> GOLInterface:
        rules = ConwayRules()
        gol = GameOfLife(0, 0, rules)
        gol.init_field(0, 0, start_generation)
        return gol

    @classmethod
    def create_from_arguments(cls, arguments: Arguments) -> GOLInterface:
        start_generation = input_field(arguments.start_generation, arguments.width)
        return cls.create_with_start_generation(start_generation)




from flask import Flask, jsonify, request
from flask_restx import Api, Resource, fields, reqparse
from flask_cors import CORS
from glife.cli.arguments import Arguments
from glife.cli.cli import create_gol_arguments
from glife.colony.colony import Colony
from glife.colony.exception.colony_exception import ColonyException
from glife.gol.factory import Factory

app = Flask(__name__)
CORS(app)
api = Api(app=app, version="0.1")
name_space = api.namespace('', description='GoL Api')

RESULT_TYPE_GENERATION_NUMBER = 'GEN_NUMBER'

MODEL_WIDTH = 'width'
MODEL_START_GENERATION = 'start'
MODEL_NUMBER_OF_GENERATIONS = 'generations'
MODEL_RESULT_TYPE = 'result'

model_colony = api.model('Colony', {
    MODEL_WIDTH: fields.Integer(),
    MODEL_START_GENERATION: fields.String,
    MODEL_NUMBER_OF_GENERATIONS: fields.Integer(default=1),
    MODEL_RESULT_TYPE: fields.String(default=RESULT_TYPE_GENERATION_NUMBER),
})


class ColonyEntity:
    def __init__(self, colony: Colony, colony_id: str):
        self.colony = colony
        self.id = colony_id

    def json(self, page: int, limit: int):
        start = (page - 1) * limit
        stop = start + limit
        return {
            'id': self.id,
            'generations': len(self.colony._field_list),
            'status': self.colony._state,
            'history': [field._field for field in self.colony._field_list[start:stop]],
            'history_page': page,
            'history_limit': limit
        }


colonies: dict = {}


def create_colony(payload) -> Colony:
    args = Arguments()
    args.start_generation = payload[MODEL_START_GENERATION]
    args.width = payload[MODEL_WIDTH]
    gol_arguments = create_gol_arguments(args)
    gol = Factory.create_from_arguments(gol_arguments)
    colony = Colony(gol)
    return colony


def process_colony(colony: Colony, generations: int) -> Colony:
    try:
        for i in range(0, generations):
            colony.next_generation()
    except ColonyException as exception:
        pass

    return colony


get_parser = reqparse.RequestParser()
get_parser.add_argument('page', type=int, default=1)
get_parser.add_argument('limit', type=int, default=20)


@name_space.route('/colonies/', defaults={'colony_id': None})
@name_space.route('/colonies/<colony_id>')
class GameOfLifeApi(Resource):

    @name_space.expect(get_parser)
    def get(self, colony_id: str):
        if colony_id:
            args = get_parser.parse_args()
            result = {}
            colony = colonies[colony_id]
            if colony:
                result = colony.json(args['page'], args['limit'])
        else:
            result = [
                {
                    'id': colony_id,
                    'status': colonies[colony_id].colony._state,
                    'generations': len(colonies[colony_id].colony._field_list)
                } for colony_id in colonies.keys()
            ]
        return jsonify(result)

    @name_space.expect(model_colony, validate=True)
    def post(self, colony_id=None):
        app.logger.debug(api.payload)

        colony = create_colony(api.payload)
        colony_id = str(colony.generate_hash(colony.field().__str__()))

        from_cache = False
        if colony_id not in colonies:
            colony = process_colony(colony, api.payload[MODEL_NUMBER_OF_GENERATIONS])
            colonies[colony_id] = ColonyEntity(colony, colony_id)
        else:
            colony = colonies[colony_id].colony
            from_cache = True

        print(f"colony hash: {colony_id}, generations: {len(colony._field_list)}, from cache: {str(from_cache)}")

        return jsonify({
            'id': colony_id,
            'generations': len(colony._field_list),
            'status': colony._state
        })


app.run(port=8080)




import pytest
from ...field import Field
from ...point import Point
from ...rules.conway import ConwayRules


@pytest.fixture
def field():
    o = Field(Point(3,3))
    return o


@pytest.fixture
def conway_rules():
    o = ConwayRules()
    return o


def test_01(field: Field, conway_rules: ConwayRules):
    field.init([
        [0, 0, 0],
        [0, 1, 0],
        [0, 0, 0],
    ])
    p = Point(1,1)
    assert field.state_point(p) == True
    if not conway_rules.calculate(field, p):
        field.set_point(p, False)
    assert field.state_point(p) == False


def test_02(field: Field, conway_rules: ConwayRules):
    field.init([
        [0, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ])
    p = Point(1, 1)
    assert field.state_point(p) == True
    assert field.state(2,2) == True
    if not conway_rules.calculate(field, p):
        field.set_point(p, False)
    assert field.state_point(p) == False


def test_03(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == True


def test_04(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 0, 0],
        [0, 1, 1],
        [0, 0, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == True


def test_05(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 0, 0],
        [1, 1, 1],
        [0, 0, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == False


def test_06(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 0, 1],
        [1, 1, 1],
        [0, 0, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == False


def test_07(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 0, 1],
        [1, 1, 0],
        [0, 1, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == False


def test_08(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == False


def test_09(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 1, 1],
        [1, 1, 1],
        [1, 1, 1],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == False


def test_10(field: Field, conway_rules: ConwayRules):
    field.init([
        [1, 1, 1],
        [1, 0, 1],
        [1, 1, 1],
    ])
    p = Point(1,1)
    assert field.state_point(p) == False
    assert conway_rules.calculate(field, p) == False


def test_11(field: Field, conway_rules: ConwayRules):
    field.init([
        [0, 0, 0],
        [1, 0, 1],
        [1, 0, 0],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == True


def test_12(field: Field, conway_rules: ConwayRules):
    field.init([
        [0, 0, 0],
        [1, 0, 1],
        [0, 0, 0],
    ])
    assert conway_rules.calculate(field, Point(1,1)) == False


def test_20(field: Field, conway_rules: ConwayRules):
    field.init([
        [0, 0, 0],
        [0, 1, 1],
        [0, 1, 1],
    ])
    assert conway_rules.calculate(field, Point(2,2)) == True




import pytest
from ..point import Point
from ..field import Field


@pytest.fixture
def field():
    o = Field(Point(10,10))
    o.set(0, 3, True)
    o.set_point(Point(7, 7), True)
    return o


def test_state(field: Field):
    assert field.state(0,0) == False
    assert field.state(0,3) == True
    assert field.state(-1,0) == False


def test_iter(field: Field):
    i = s = 0
    points = []
    for p in field:
        i += 1
        if field.state_point(p):
            s += 1
            points.append(p)
    assert i == 100
    assert s == 2
    assert len(points) == 2
    assert points[0].x == 0
    assert points[0].y == 3
    assert points[1].x == 7
    assert points[1].y == 7


def test_all_on(field: Field):
    for p in field:
        field.set_point(p, True)
    assert field.state(0, 0) == True
    assert field.state(9, 9) == True
    assert field.state(0, 10) == False
    assert field.state(10, 0) == False



import argparse


class Arguments:
    width: int = 0
    height: int = 0
    start_generation: str = ''
    file_start_generation: str = ''
    file_false: str = '0'
    file_true: str = '1'
    generations: int = 1
    interactive: bool = False
    clear: bool = False
    quiet: bool = False


def parse() -> object:
    parser = argparse.ArgumentParser(
        epilog='2021 Łukasz Bacik <mail@luka.sh> https://github.com/lbacik/glife'
    )

    parser.add_argument('-s', '--start-generation', nargs='?', default=Arguments.start_generation)
    parser.add_argument('-f', '--file-start-generation', nargs='?', default=Arguments.file_start_generation)
    parser.add_argument('--file-false', nargs='?', default=Arguments.file_false)
    parser.add_argument('--file-true', nargs='?', default=Arguments.file_true)
    parser.add_argument('-w', '--width', nargs='?', default=Arguments.width, type=int)
    parser.add_argument('-g', '--generations', nargs='?', default=Arguments.generations, type=int)
    parser.add_argument('-i', '--interactive', action='store_const', const=True)
    parser.add_argument('-c', '--clear', action='store_const', const=True)
    parser.add_argument('-q', '--quiet', action='store_const', const=True)

    return parser.parse_args(namespace=Arguments)
